@@ -12,7 +12,7 @@ from pairtree import FileNotFoundException
 
 from datetime import datetime
 
-from os import path, mkdir
+from os import path, mkdir, rename
 
 from shutil import copy2
 
@@ -53,6 +53,16 @@ class HarvestedRecord(object):
         self.versions=None
         self.currentversion=None
     
+    def __setattribute__(self,name, value):
+        if name not in ['files', 'versions', 'currentversion', 'date']:
+            if name == "metadata":
+                self.manifest['metadata'] = value
+            else:
+                return object.__setattribute__(self, name, value)
+        else:
+            logger.error("Attribute %s cannot be set in this manner" % name)
+            raise Exception
+    
     def __getattribute__(self,name):
         if name=='files':
             return self.manifest['files'][self.manifest['currentversion']]
@@ -60,6 +70,12 @@ class HarvestedRecord(object):
             return self.manifest['versions']
         elif name=='currentversion':
             return self.manifest['currentversion']
+        elif name=='date':
+            return self.manifest['date']
+        elif name=='metadata':
+            if not self.manifest.has_key('metadata'):
+                self.manifest['metadata'] = {}
+            return self.manifest['metadata']
         else:
             return object.__getattribute__(self, name)
 
@@ -75,26 +91,35 @@ class HarvestedRecord(object):
         if version not in self.manifest['versions']:
             self.manifest['versions'].append(version)
         self.manifest['version_dates'][version] = date
+        self.manifest['subdir'][version] = []
         self.manifest['metadata_files'][version] = []
         self.manifest['files'][version] = []
         self.set_version_date(version, date)
-        self.po.add_bytestream_by_path(path.join(str(version), "4=%s" % id_encode(self.item_id)), self.item_id)
+        self.po.add_bytestream_by_path(path.join("__"+str(version), "4=%s" % id_encode(self.item_id)), self.item_id)
 
     def _init_manifests_emptydatastructures(self):
         self.item_id = self.manifest['item_id']
         if not self.manifest.has_key('metadata_files'):
             self.manifest['metadata_files'] = {}
+        if not self.manifest.has_key('subdir'):
+            self.manifest['subdir'] = {}
         if not self.manifest.has_key('files'):
             self.manifest['files'] = {}
         for version in self.manifest['versions']:
             if not self.manifest['metadata_files'].has_key(version):
                 self.manifest['metadata_files'][version] = []
+            if not self.manifest['subdir'].has_key(version):
+                self.manifest['subdir'][version] = []
 
     def _reload_filelist(self, version):
         if self.manifest['files'].has_key(version) and version in self.manifest['versions']:
             self.manifest['files'][version] = []
+            self.manifest['subdir'][version] = []
             # init from disc
-            for filename in [x for x in self.po.list_parts(version) if NAMASTE_PATTERN.match(x) != None]:
+            for filename in [x for x in self.po.list_parts("__"+str(version)) if not (x.startswith("0=") or x.startswith("1=") or x.startswith("2=") or x.startswith("3=") or x.startswith("4=") or x.startswith("5=")) ]:
+                logger.debug("Item %s has file: %s" % (self.item_id, x) )
+                if self.po.isdir(path.join("__"+str(version), filename)):
+                    self.manifest['subdir'][version].append(filename)
                 self.manifest['files'][version].append(filename)
 
     def _init_manifest(self):
@@ -104,6 +129,7 @@ class HarvestedRecord(object):
         self.manifest['item_id'] = self.po.id
         self.manifest['versions'] = []
         self.manifest['version_dates'] = {}
+        self.manifest['subdir'] = {}
         self.manifest['currentversion'] = "1"
         self._setup_version_dir("1", self.manifest['date'])
         self.manifest.sync()
@@ -111,7 +137,7 @@ class HarvestedRecord(object):
     def _read_date(self, version = None):
         if not version:
             version = self.manifest['currentversion']
-        date_namaste_tags = [x for x in self.po.list_parts(str(version)) if x.startswith("3=")]
+        date_namaste_tags = [x for x in self.po.list_parts("__"+str(version)) if x.startswith("3=")]
         if len(date_namaste_tags) >= 1:
             lmd = date_namaste_tags.pop()[2:]   # take the first tag and remove the '3='
             lmd = id_decode(lmd)                # reverse the 'pairtree' encoding of the date
@@ -130,11 +156,11 @@ class HarvestedRecord(object):
                 metadata = False
                 if x in self.manifest['metadata_files'][latest_version]:
                     metadata = True
-                self.add_stream(x, stream, metadata=metadata)
+                self.put_stream(x, stream, metadata=metadata)
 
     def path_to_item(self):
         return self.po.fs._id_to_dirpath(self.po.id)
-    
+
     def revert(self, **kw):
         if not path.isdir(self.itempath):
             logger.error("Path to harvested item does not exist")
@@ -177,20 +203,24 @@ class HarvestedRecord(object):
         self.manifest.revert()
         self._init_manifests_emptydatastructures()
     
-    def add_stream(self, filename, stream, metadata=True):
-        if metadata and filename not in self.manifest['metadata_files'][self.manifest['currentversion']]:
-            self.manifest['metadata_files'][self.manifest['currentversion']].append(filename)
-        if filename not in self.manifest['files'][self.manifest['currentversion']]:
-            self.manifest['files'][self.manifest['currentversion']].append(filename)
+    def put_stream(self, filename, stream, metadata=False):
+        version = self.manifest['currentversion']
+        if metadata and filename not in self.manifest['metadata_files'][version]:
+            self.manifest['metadata_files'][version].append(filename)
+        if filename not in self.manifest['files'][version]:
+            self.manifest['files'][version].append(filename)
+        self.po.add_bytestream_by_path(path.join("__" + str(version), filename), stream)
+        self._reload_filelist(version)
         self.sync()
-        return self.po.add_bytestream_by_path(path.join(str(self.manifest['currentversion']), filename), stream)
-    
-    def get_stream(self, filename, version=None):
+        return 
+    def get_stream(self, filename, version=None, writeable=False):
+        """NB If writeable is set to True, then the file is opened "wb+" and can accept writes.
+        Otherwise, the file is opened read-only."""
         if not version:
             version = self.manifest['currentversion']
-        if filename in self.manifest['files'][version]:
-            return self.po.get_bytestream_by_path(path.join(str(version), filename),
-                                                  streamable=True)
+        if self.isfile(filename, version):
+            return self.po.get_bytestream_by_path(path.join("__" + str(version), filename),
+                                                  streamable=True, appendable=writeable)
         else:
             raise FileNotFoundException
     
@@ -199,14 +229,28 @@ class HarvestedRecord(object):
             versions = [self.manifest['currentversion']]
         for version in versions:
             try:
-                self.po.del_file_by_path(path.join(str(version), filename))
-                if filename in self.manifest['metadata_files'][version]: 
+                self.po.del_file_by_path(path.join("__" + str(version), filename))
+                if self.isfile(filename, version):
                     self.manifest['metadata_files'][version].remove(filename)
                 self._reload_filelist(version)
             except FileNotFoundException:
                 logger.info("File %s not found at version %s and so cannot be deleted" % (filename, version))
         self.sync()
 
+    def isfile(self, filepath, version=None):
+        if not version:
+            version = self.manifest['currentversion']
+        return self.po.isfile(path.join("__"+str(version), filepath))
+
+    def isdir(self, filepath, version=None):
+        if not version:
+            version = self.manifest['currentversion']
+        return self.po.isdir(path.join("__"+str(version),filepath))
+
+    def list_parts(self, subpath):
+        if self.isdir(subpath):
+            return self.po.list_parts(path.join("__"+str(self.manifest['currentversion']),subpath))
+
     def set_version_cursor(self, version):
         if version in self.manifest['versions']:
             self.manifest['currentversion'] = version
@@ -246,6 +290,24 @@ class HarvestedRecord(object):
             logger.error("Version %s is not found in the object. Cannot be cloned" % original_version)
             return False
 
+    def rename_version(self, original_version, new_name):
+        if original_version in self.manifest['versions']:
+            self.manifest['versions'].append(new_name)
+            self.manifest['version_dates'][new_name] = self.manifest['version_dates'][original_version]
+            self.manifest['files'][new_name] = self.manifest['files'][original_version]
+            self.manifest['metadata_files'][new_name] = self.manifest['metadata_files'][original_version]
+            rename(path.join(self.path_to_item(), "__"+str(original_version)), path.join(self.path_to_item(), "__" + str(new_name)))
+            self.set_version_cursor(new_name)
+            self.manifest['versions'].remove(original_version)
+            del self.manifest['version_dates'][original_version]
+            del self.manifest['files'][original_version]
+            del self.manifest['metadata_files'][original_version]
+            self.sync()
+            return new_name
+        else:
+            logger.error("Version %s is not found in the object. Cannot be renamed" % original_version)
+            return False
+
     def create_new_version(self, version, date=None):
         version = str(version)
         if version not in self.manifest['versions']:
@@ -274,7 +336,7 @@ class HarvestedRecord(object):
                 self._read_date()
                 self._setup_version_dir(version, date)
             self.sync()
-            return self.po.del_path(version, recursive=True)
+            return self.po.del_path("__"+str(version), recursive=True)
 
     def del_versions(self, versions=[]):
         results = []
@@ -285,7 +347,7 @@ class HarvestedRecord(object):
     def set_version_date(self, version, date):
         if version in self.manifest['versions']:
             self.manifest['version_dates'][version] = date
-            self.po.add_bytestream_by_path(path.join(str(version), "3=%s" % id_encode(date)), date)
+            self.po.add_bytestream_by_path(path.join("__"+str(version), "3=%s" % id_encode(date)), date)
             return True
         else:
             logger.error("Version %s does not exist" % version)
@@ -319,14 +381,38 @@ class Silo(object):
             logger.error("Cannot setup the state persistence file at %s/%s" % (self.state['storage_dir'], PERSISTENCE_FILENAME))
             raise Exception("Cannot setup the state persistence file at %s/%s" % (self.state['storage_dir'], PERSISTENCE_FILENAME))
 
+    def __iter__(self):
+        return self.list_items()
+
+    def iteritems(self):
+        for item in self.list_items():
+            yield self.get_item(item)
+
+    def __getitem__(self, key):
+        if self.exists(key):
+            return self.get_item(key)
+
+    def __len__(self): return len(self.keys())
+
+    def keys(self): return [x for x in self.__iter__()]
+    def has_key(self, key): return self.exists(key)
     def exists(self, item_id):
         return self._store.exists(item_id)
 
-    def get_item(self, item_id):
-        p_obj = self._store.get_object(item_id)
-        return HarvestedRecord(p_obj)
+    def get_item(self, item_id, date=None, force=False):
+        if self.exists(item_id):
+            p_obj = self._store.get_object(item_id)
+            return HarvestedRecord(p_obj, date)
+        elif self.exists(self.state['uri_base'] + item_id) and not force:
+            p_obj = self._store.get_object(self.state['uri_base'] + item_id)
+            return HarvestedRecord(p_obj, date)
+        else:
+            p_obj = self._store.get_object(item_id)
+            return HarvestedRecord(p_obj, date)
 
     def del_item(self, item_id):
         return self._store.delete_object(item_id)
 
+    def list_items(self):
+        return self._store.list_ids()
 
